;- QBPlay v1.5
;- 2007/09/08 by Froggerprogger
;/
;/ QBPlay() interpretates a String similar to those of QuickBasic (but with even more
;/ functionality) and creates WAV-Data and a 'Sound' from it.
;/ You might just play it via PB's PlaySound() (or the wrapper-procedure QBPlay_Start())
;/ or mix two songs together using the QBPlay_Mix()
;/ or save your song to a WAV-file.
;/
;/ EBNF for the PlayString (xx stands for any integer):
;/ ( [(a|b|c|d|e|f|g|h)[#|+|-][xx][.]] | Oxx | < | > | Nxx | Txx |  | Lxx | P[xx] | M[L|N|S|xx] | X(V|P|S|A)xx)*
;/
;/  a,b,c,d,e,f,h   = note to play (h = b)
;/  a[#,+,-][xx][.] = note 'a' (same for b,c...) to play of length xx, xx is integer from 1 to 64
;/     # or +       = increase last note by one halftone (you might use multiple # or +)
;/     -            = decrease last note by one halftone (you might use multiple -)
;/     .            = add half of the length to the last note (you might use multiple .)
;/  Oxx             = change octave for all following notes, xx is integer from 0-7, just 'O' will set to octave 0 ('O' = octave)
;/  < or >          = changes octave by +1 or -1 for all following notes (you might use multiple < or >)
;/  Nxx             = note to play given by it's ID (0 = pause, 1-96 = note-ID), just 'N' will play a pause ('N' = number)
;/  Txx             = sets the tempo in bpm, xx is integer from 10-360 ('T' = tempo)
;/  Lxx             = length for all following notes, xx is integer from 1-64 ('L' = length)
;/  P               = play a pause with the current length for notes ('P' = pause)
;/  Pxx             = plays a pause with the given length. xx is integer from 1-64 ('P' = pause)
;/  Mxx             = set legato-length of notes in percent (1 = very staccato, 100 = total legato)
;/  ML              = play legato (value 100) ('L' = legato)
;/  MN              = play normal (value 80) ('N' = normal)
;/  MS              = play staccato (value 20) ('S' = staccato)
;/  XVxx            = sets notes' in percent, xx is integer >= 0 ('V' = volume)
;/  XPxx            = sets Pan, xx is integer from -100(left) to 100 (right) ('P' = pan)
;/  XIxx            = sets instrument, id is from 0 to 4 ('I' = instrument)
;/  XSxx            = sustain ('S' = sustain)
;/  XAxx            = attack ('A' = attack)
;/
;/ pan-values are always given from -100 (left) to 100 (right)
;/ volume-values are always given in percent from 0 (no sound) over 100 (normal) to endless amplify
;/ mix-volume-values are always given from -100 (first song) to 100 (second song)
;/ attack/sustain-values are given in milliseconds (keep in mind: the longer the sustain, the longer the calculation lasts!)
;/ instruments are set by their id (or the related constant)
;/ playmode is 0=play once, 1=play looped, 2=start paused
;/
;/ Have fun with it !
;/ please tell me about bugs or wishes
;/
;/ kurras@schalldesign.de


IncludeFile "QBPlay_1.5_RESIDENT.pb" ; exclude, if you have this resident-file installed


;- necessary procedure-declarations
Declare.l QBPlay_CreateSong(*p_song.QBSong, p_samplerate.l, p_bitrate.l)
Declare.l QBPlay_Parse(*p_song.QBSong, p_samplerate.l, p_bitrate.l, p_write.l)
Declare.l QBPlay_WriteNote(*lpActSample, p_fq.f, noteOnLengthSample.l, *p_WAVfmtEx.WAVEFORMATEX, *p_tone.QBSong_ToneData, p_mode)
DeclareDLL.l QBPlay4(p_str.s, p_playmode.l, p_volume.l, p_pan.l, p_soundID.l, p_attack.l, p_sustain.l)
DeclareDLL.l QBPlay_Free2(*p_song.QBSong, p_keepSound)



;- the procedures
;/ QBPlay() - main procedure
ProcedureDLL.l QBPlay(p_str.s) ; creates a new QBsong and sound given by the string p_str
  ProcedureReturn QBPlay4(p_str, 0, 100, 0, 0, 25, 25)
EndProcedure
ProcedureDLL.l QBPlay2(p_str.s, p_playmode.l)
  ProcedureReturn QBPlay4(p_str, p_playmode, 100, 0, 0, 25, 25)
EndProcedure
ProcedureDLL.l QBPlay3(p_str.s, p_playmode.l, p_volume.l, p_pan.l)
  ProcedureReturn QBPlay4(p_str, p_playmode, p_volume, p_pan, 0, 25, 25)
EndProcedure
ProcedureDLL.l QBPlay4(p_str.s, p_playmode.l, p_volume.l, p_pan.l, p_soundID.l, p_attack.l, p_sustain.l)
  Protected *song.QBSong

  If p_str = "" : ProcedureReturn 0 : EndIf
  If p_playmode < 0 Or p_playmode > 2 : ProcedureReturn 0 : EndIf
  If p_pan < -100 : p_pan = -100 : EndIf
  If p_pan > 100  : p_pan = 100  : EndIf
  If p_volume < 0   : p_volume = 0 : EndIf
  If p_attack < 0 Or p_sustain < 0 Or p_soundID < 0
    ProcedureReturn 0
  EndIf
  
  *song = AllocateMemory(SizeOf(QBSong))

  If *song = 0 : ProcedureReturn 0 : EndIf
  
  p_str = UCase(p_str)
  p_str = ReplaceString(p_str, " ", "")
  p_str = ReplaceString(p_str, "#", "+")
  p_str = ReplaceString(p_str, "H", "B")

  ; set the defaults
  *song\parse\playStr           = p_str
  *song\parse\playStrLen        = Len(p_str)
  *song\parse\actPos            = 1
  *song\parse\tempo             = 120
  *song\parse\octave            = 3
  *song\parse\Length            = 4
  *song\parse\legato            = 80
  *song\tone\volume             = p_volume
  *song\tone\pan                = p_pan
  *song\tone\id                 = p_soundID
  *song\tone\attack             = p_attack
  *song\tone\sustain            = p_sustain

  If QBPlay_CreateSong(*song, #QBPlay_SampleRate, #QBPlay_BitRate) = 0
    FreeMemory(*song)
    ProcedureReturn 0
  EndIf

  *song\lpSound = CatchSound(#PB_Any, *song\lpWAV)
  If *song\lpSound = 0
    FreeMemory(*song\lpWAV)
    FreeMemory(*song)
    ProcedureReturn 0
  EndIf
  
  If p_playmode = 0
    PlaySound(*song\lpSound, 0)
    ProcedureReturn *song
  ElseIf p_playmode = 1
    PlaySound(*song\lpSound, 1)
    ProcedureReturn *song
  ElseIf p_playmode = 2
    ProcedureReturn *song
  EndIf
EndProcedure

;/ procedures for information/navigation
ProcedureDLL.l QBPlay_Start(*p_song.QBSong, playmode.l) ; starts playing a QBSong
  ProcedureReturn PlaySound(*p_song\lpSound, playmode)
EndProcedure
ProcedureDLL.l QBPlay_Stop(*p_song.QBSong, playmode.l)  ; stops playing a QBSong
  ProcedureReturn StopSound(*p_song\lpSound)
EndProcedure
ProcedureDLL.l QBPlay_GetSound(*p_song.QBSong) ; returns handle to QBSong's sound for direct usage with PlaySound(), etc.
  ProcedureReturn *p_song\lpSound
EndProcedure
ProcedureDLL.l QBPlay_GetLengthMs(*p_song.QBSong) ; returns the length in ms of the given QBSong
  ProcedureReturn *p_song\lengthMs
EndProcedure

;/ procedures on songs
ProcedureDLL.l QBPlay_Copy(*p_song.QBSong) ; Creates a copy of the given *p_song and returns a pointer to it
  Protected *songCpy.QBSong
  Protected *WAVfmtEx.WAVEFORMATEX
  
  If *p_song = 0 : ProcedureReturn 0 : EndIf

  *WAVfmtEx = *p_song\lpWAV + 20

  *songCpy = AllocateMemory(SizeOf(QBSong))
  If *songCpy = 0 : ProcedureReturn 0 : EndIf
  
  CopyMemory(*p_song, *songCpy, SizeOf(QBSong))
  
  *songCpy\lpWAV = AllocateMemory(28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx\cbSize + *p_song\WAVdataLen)
  If *songCpy\lpWAV = 0
    FreeMemory(*songCpy)
    ProcedureReturn 0
  EndIf

  CopyMemory(*p_song\lpWAV, *songCpy\lpWAV, 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx\cbSize + *p_song\WAVdataLen)
  
  *songCpy\lpSound = CatchSound(#PB_Any, *songCpy\lpWAV)
  If *songCpy\lpSound = 0
    FreeMemory(*songCpy\lpWAV)
    FreeMemory(*songCpy)
    ProcedureReturn 0
  EndIf
  
  ProcedureReturn *songCpy
EndProcedure
ProcedureDLL.l QBPlay_Mix(*p_song1.QBSong, *p_song2.QBSong, p_mix.l) ; mixes two songs and stores the result in *p_song1. p_mix is from -100 (*p_song1 only) to 100 (*p_song2 only) and gives the volume-balance.
  Protected *WAVfmtEx1.WAVEFORMATEX, *WAVfmtEx2.WAVEFORMATEX
  Protected temp.l
  
  If *p_song1 = 0 Or *p_song2 = 0 : ProcedureReturn 0 : EndIf
  If p_mix < -100 : p_mix = -100 : EndIf
  If p_mix > 100  : p_mix = 100  : EndIf
  
  *WAVfmtEx1 = *p_song1\lpWAV + 20
  *WAVfmtEx2 = *p_song2\lpWAV + 20
  
  If *WAVfmtEx1\wBitsPerSample <> 16 Or *WAVfmtEx2\wBitsPerSample <> 16 Or *WAVfmtEx1\nSamplesPerSec <> *WAVfmtEx2\nSamplesPerSec
    ProcedureReturn 0
  EndIf

  If *p_song2\WAVdataLen > *p_song1\WAVdataLen
    temp = ReAllocateMemory(*p_song1\lpWAV, 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx1\cbSize + *p_song2\WAVdataLen)
    If temp = 0
      ProcedureReturn 0
    EndIf
    *p_song1\lpWAV  = temp
    *WAVfmtEx1      = *p_song1\lpWAV + 20
    
    *p_song1\WAVdataLen = *p_song2\WAVdataLen
    PokeL(*p_song1\lpWAV + 4  , 20 + SizeOf(WAVEFORMATEX) + *WAVfmtEx1\cbSize + *p_song1\WAVdataLen) ; size of RIFF-chunk
    PokeL(*p_song1\lpWAV + 24 + SizeOf(WAVEFORMATEX) + *WAVfmtEx1\cbSize, *p_song1\WAVdataLen) ; size of data-chunk
    
    *p_song1\lengthMs = *p_song2\lengthMs
  EndIf

  Protected resume.l, i1.l, i2.l
  Protected *sampleVal1.WORD, *sampleVal2.WORD
  Protected s1.l, s2.l
  
  resume = 1
  *sampleVal1 = *p_song1\lpWAV + 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx1\cbSize
  *sampleVal2 = *p_song2\lpWAV + 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx2\cbSize
  While resume
    If i1 < *p_song1\WAVdataLen 
      s1 = *sampleVal1\w
    Else
      s1 = 0
    EndIf
    
    If i2 < *p_song2\WAVdataLen 
      s2 = *sampleVal2\w
    Else
      s2 = 0
    EndIf
    
    If p_mix = 0
      *sampleVal1\w = (s1 + s2) / 2
    ElseIf p_mix < 0
      *sampleVal1\w = (s1 * (0 - p_mix) + s2 * (100 + p_mix)) / 100
    Else
      *sampleVal1\w = (s1 * (100 - p_mix) + s2 * p_mix) / 100
    EndIf
    
    i1 + 2 : *sampleVal1 + 2
    i2 + 2 : *sampleVal2 + 2
    If i1 > *p_song1\WAVdataLen And i2 > *p_song2\WAVdataLen
      resume = 0
    EndIf
  Wend

  If *p_song1\lpSound
    StopSound(*p_song1\lpSound)
    FreeSound(*p_song1\lpSound)
    *p_song1\lpSound = CatchSound(#PB_Any, *p_song1\lpWAV)
    If *p_song1\lpSound = 0
      ProcedureReturn 0
    EndIf
  EndIf
  
  ProcedureReturn *p_song1
EndProcedure
ProcedureDLL.l QBPlay_WriteToFile(*p_song.QBSong, p_filename.s) ; creates a file named p_filename and stores the song in WAV-format in there
  Protected hFile.l
  Protected *WAVfmtEx.WAVEFORMATEX
  
  If p_filename = "" Or *p_song = 0 : ProcedureReturn 0 : EndIf
  
  *WAVfmtEx = *p_song\lpWAV + 20
   
  hFile = CreateFile(#PB_Any, p_filename)
  If hFile = 0 : ProcedureReturn 0 : EndIf
  
  WriteData(hFile, *p_song\lpWAV, 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx\cbSize + *p_song\WAVdataLen)
  
  CloseFile(hFile)
  ProcedureReturn 1
EndProcedure
ProcedureDLL.l QBPlay_Free(*p_song.QBSong) ; frees the QBSong and the created sound. (set p_keepSound to keep the sound and just delete the QBSong)
  ProcedureReturn QBPlay_Free2(*p_song, 0)
EndProcedure
ProcedureDLL.l QBPlay_Free2(*p_song.QBSong, p_keepSound)
  If *p_song = 0 : ProcedureReturn 0 : EndIf
  FreeMemory(*p_song\lpWAV)
  If p_keepSound = 0
    StopSound(*p_song\lpSound)
    FreeSound(*p_song\lpSound)
  EndIf
  FreeMemory(*p_song)
  ProcedureReturn 1
EndProcedure

;/ procedures for internal usage only
Procedure.l QBPlay_CreateSong(*p_song.QBSong, p_samplerate.l, p_bitrate.l)
  Protected songlengthSamples.l
  Protected WAVfmtEx.WAVEFORMATEX
  Protected songDefData.QBSong
  
  CopyMemory(*p_song, @songDefData, SizeOf(QBSong))
  songlengthSamples = QBPlay_Parse(*p_song, p_samplerate, p_bitrate, 0)
  If songlengthSamples <= 0
    ProcedureReturn 0
  EndIf
  CopyMemory(@songDefData, *p_song, SizeOf(QBSong))

  WAVfmtEx\wFormatTag        = #WAVE_FORMAT_PCM
  WAVfmtEx\nChannels         = 2
  WAVfmtEx\nSamplesPerSec    = p_samplerate
  WAVfmtEx\wBitsPerSample    = p_bitrate
  WAVfmtEx\nBlockAlign       = Round(WAVfmtEx\wBitsPerSample / 8, 1) * WAVfmtEx\nChannels
  WAVfmtEx\nAvgBytesPerSec   = WAVfmtEx\nSamplesPerSec * WAVfmtEx\nBlockAlign
  WAVfmtEx\cbSize            = 0

  *p_song\WAVdataLen         = songlengthSamples * WAVfmtEx\nBlockAlign

  *p_song\lpWAV = AllocateMemory(28 + SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize + *p_song\WAVdataLen)
  If *p_song\lpWAV = 0
    ProcedureReturn 0
  EndIf

  PokeL(*p_song\lpWAV       , 'FFIR') ; id RIFF-chunk
  PokeL(*p_song\lpWAV + 4   , 20 + SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize + *p_song\WAVdataLen) ; size of RIFF-chunk
  PokeL(*p_song\lpWAV + 8   , 'EVAW') ; id WAVE-chunk
  PokeL(*p_song\lpWAV + 12  , ' tmf') ; id fmt-chunk
  PokeL(*p_song\lpWAV + 16  , SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize) ; size fmt-chunk (incl. opt. extra-fmt-Data-length-word here)
  CopyMemory(@WAVfmtEx, *p_song\lpWAV + 20, SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize)
  PokeL(*p_song\lpWAV + 20 + SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize, 'atad') ; id data-chunk
  PokeL(*p_song\lpWAV + 24 + SizeOf(WAVEFORMATEX) + WAVfmtEx\cbSize, *p_song\WAVdataLen) ; size of data-chunk

  If QBPlay_Parse(*p_song, p_samplerate, p_bitrate, 1) = 0
    FreeMemory(*p_song\lpWAV)
    *p_song\lpWAV = 0
    ProcedureReturn 0
  EndIf

  *p_song\lengthMs = songlengthSamples / 44.1
  
  ProcedureReturn 1
EndProcedure
Procedure.l QBPlay_Parse(*p_song.QBSong, p_samplerate.l, p_bitrate.l, p_mode.l)
  Protected *char.BYTE
  Protected temp.l, tempF.f, i.l
  Protected parseToNumEnd.l, parseToIncDecEnd.l, parseToDoubleEnd.l
  Protected noteID.l, noteFQ.f, noteLength.f, noteOnLengthSample.l, noteLengthSample.l
  Protected actSample.l, *lpActSample.l, maxWrittenSample.l, maxmaxWrittenSample.l
  Protected *lpDataOffset.l, *WAVfmtEx.WAVEFORMATEX
  Protected resume.l
  Protected lastDelayTime.l

  If *p_song = 0 : ProcedureReturn 0 : EndIf

  If p_mode
    *WAVfmtEx = *p_song\lpWAV + 20
    If *WAVfmtEx = 0 : ProcedureReturn 0 : EndIf
    *lpDataOffset = *p_song\lpWAV + 28 + SizeOf(WAVEFORMATEX) + *WAVfmtEx\cbSize
  EndIf
  
  *p_song\parse\actPos = 1
  
  *char               = @*p_song\parse\playStr
  noteID              = 0
  actSample           = 0
  resume              = 1
  totalSamples        = 0
  noteOnLengthSample  = 0
  noteLengthSample    = 0
  noteFQ              = 0.0

  While resume
    playNow = 0

    While *p_song\parse\actPos <= *p_song\parse\playStrLen And playNow = 0
      noteLength        = *p_song\parse\Length
      parseToIncDecEnd  = 0
      parseToNumEnd     = 0
      parseToDoubleEnd  = 0
      Select *char\b & $FF
        Case 'T' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 1, *p_song\parse\playStrLen - *p_song\parse\actPos - 1))
                    If temp = 0     : temp = 120  : EndIf
                    If temp < 10    : temp = 10   : EndIf
                    If temp > 360   : temp = 360  : EndIf
                    *p_song\parse\tempo   = temp
                    parseToNumEnd         = 1
        
        Case 'N' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 1, *p_song\parse\playStrLen - *p_song\parse\actPos - 1))
                    If temp < 0     : temp = 0    : EndIf
                    If temp > 96    : temp = 96   : EndIf
                    noteID          = temp
                    parseToNumEnd   = 1
                    playNow         = 1
        
        Case 'L' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 1, *p_song\parse\playStrLen - *p_song\parse\actPos - 1))
                    If temp < 1     : temp = 1    : EndIf
                    If temp > 64    : temp = 64   : EndIf
                    *p_song\parse\Length  = temp
                    noteLength            = temp
                    parseToNumEnd         = 1
  
        Case 'O' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 1, *p_song\parse\playStrLen - *p_song\parse\actPos - 1))
                    If temp < 0     : temp = 0   : EndIf
                    If temp > 7     : temp = 7   : EndIf
                    *p_song\parse\octave  = temp
                    parseToNumEnd   = 1
        
        Case 'P' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 1, *p_song\parse\playStrLen - *p_song\parse\actPos - 1))
                    If temp < 1     : temp = *p_song\parse\Length  : EndIf
                    If temp > 64    : temp = 64   : EndIf
                    noteID      = 0
                    noteLength      = temp
                    parseToNumEnd   = 1
                    playNow         = 1
        
        Case '<' :  If *p_song\parse\octave < 7 : *p_song\parse\octave + 1 : EndIf
        Case '>' :  If *p_song\parse\octave > 0 : *p_song\parse\octave - 1 : EndIf
        
        Case 'A' :  noteID = 1 + *p_song\parse\octave * 12 + 9  : parseToIncDecEnd = 1 : playNow = 1
        Case 'B' :  noteID = 1 + *p_song\parse\octave * 12 + 11 : parseToIncDecEnd = 1 : playNow = 1
        Case 'C' :  noteID = 1 + *p_song\parse\octave * 12      : parseToIncDecEnd = 1 : playNow = 1
        Case 'D' :  noteID = 1 + *p_song\parse\octave * 12 + 2  : parseToIncDecEnd = 1 : playNow = 1
        Case 'E' :  noteID = 1 + *p_song\parse\octave * 12 + 4  : parseToIncDecEnd = 1 : playNow = 1
        Case 'F' :  noteID = 1 + *p_song\parse\octave * 12 + 5  : parseToIncDecEnd = 1 : playNow = 1
        Case 'G' :  noteID = 1 + *p_song\parse\octave * 12 + 7  : parseToIncDecEnd = 1 : playNow = 1
        
        Case 'M' :  *char + 1
                    *p_song\parse\actPos + 1
                    If *char\b & $FF = 'L'
                      *p_song\parse\legato = 100
                    ElseIf *char\b & $FF = 'N'
                      *p_song\parse\legato = 80
                    ElseIf *char\b & $FF = 'S'
                      *p_song\parse\legato = 20
                    Else
                      temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos, *p_song\parse\playStrLen - *p_song\parse\actPos))
                      If temp < 1     : temp = 80    : EndIf
                      If temp > 100   : temp = 100   : EndIf
                      *p_song\parse\legato = temp
                      parseToNumEnd   = 1
                      *char - 1
                      *p_song\parse\actPos - 1
                    EndIf
                    
        Case 'X' :  temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos + 2, *p_song\parse\playStrLen - *p_song\parse\actPos - 2))
                    *char + 1
                    *p_song\parse\actPos + 1
                    If *char\b & $FF = 'V'
                      If temp < 0   : temp = 0   : EndIf
                      *p_song\tone\volume = temp
                    ElseIf *char\b & $FF = 'P'
                      If temp < -100   : temp = -100  : EndIf
                      If temp > 100    : temp = 100   : EndIf
                      *p_song\tone\pan = temp
                    ElseIf *char\b & $FF = 'I'
                      If temp < 0   : temp = 0   : EndIf
                      If temp > #QBPlay_MaxInstrumentID : temp = #QBPlay_MaxInstrumentID : EndIf
                      *p_song\tone\id = temp
                    ElseIf *char\b & $FF = 'S'
                      If temp < 0   : temp = 0   : EndIf
                      *p_song\tone\sustain = temp
                    ElseIf *char\b & $FF = 'A'
                      If temp < 0   : temp = 0   : EndIf
                      *p_song\tone\attack = temp
                    Else
                      *char - 1
                      *p_song\parse\actPos - 1                      
                    EndIf
                    parseToNumEnd = 1
           
      EndSelect

      If parseToIncDecEnd ; parse all following '+' or '-'
        *char + 1
        *p_song\parse\actPos + 1
        While *p_song\parse\actPos <= *p_song\parse\playStrLen And (*char\b & $FF = '+' Or *char\b & $FF = '-')
          If *char\b & $FF = '+'
            noteID + 1
          Else
            noteID - 1
          EndIf
          *char + 1
          *p_song\parse\actPos + 1
        Wend
        If noteID < 1  : noteID = 1  : EndIf
        If noteID > 96 : noteID = 96 : EndIf
        
        ; now parse for (optional) following length
        If *char\b & $FF < '0' Or *char\b & $FF > '9'
          parseToNumEnd = 0
        Else
          temp = Val(Mid(*p_song\parse\playStr, *p_song\parse\actPos, *p_song\parse\playStrLen - *p_song\parse\actPos))
          If temp <= 0 Or temp > 64 : temp = *p_song\parse\length : EndIf
          noteLength    = temp
          parseToNumEnd = 1
        EndIf
        
        parseToDoubleEnd  = 1
      EndIf

      If parseToNumEnd ; parse to end of a number
        *char + 1
        *p_song\parse\actPos + 1
        While *p_song\parse\actPos <= *p_song\parse\playStrLen And *char\b & $FF >= '0' And *char\b & $FF <= '9' 
          *char + 1
          *p_song\parse\actPos + 1
        Wend
      EndIf
      
      If parseToDoubleEnd ; parse all following '.'
        While *p_song\parse\actPos <= *p_song\parse\playStrLen And *char\b & $FF = '.'
          noteLength * 2.0 / 3.0
          *char + 1
          *p_song\parse\actPos + 1
        Wend
        If noteLength < 1.0 : noteLength = 1.0 : EndIf
      EndIf

      If parseToIncDecEnd = 0 And parseToNumEnd = 0 And parseToDoubleEnd = 0
        *char + 1
        *p_song\parse\actPos + 1
      EndIf
    Wend

    noteLengthSample = (240 * p_samplerate) / (*p_song\parse\tempo * noteLength)
    noteOnLengthSample = (*p_song\parse\legato * noteLengthSample) / 100
    
    If p_mode ;/ write the note
      If noteID
        noteFQ = #C0 * Pow(2.0, (noteID - 1.0) / 12.0)
      Else
        noteFQ = 0
      EndIf

      *lpActSample = *lpDataOffset + actSample * *WAVfmtEx\nBlockAlign
      
      noteOnLengthSample = QBPlay_WriteNote(*lpActSample, noteFQ, noteOnLengthSample, *WAVfmtEx, @*p_song\tone, 1)
      ;now noteOnLengthSample holds the n.o. samples incl. sustain-tail
      maxWrittenSample = actSample + noteOnLengthSample
      If maxWrittenSample > maxmaxWrittenSample
        maxmaxWrittenSample = maxWrittenSample
      EndIf
    Else
      maxWrittenSample = actSample + noteOnLengthSample + (*p_song\tone\sustain * p_samplerate / 1000)
      If maxWrittenSample > maxmaxWrittenSample
        maxmaxWrittenSample = maxWrittenSample
      EndIf
    EndIf
    
    actSample + noteLengthSample
    
    If *p_song\parse\actPos > *p_song\parse\playStrLen
      resume = 0
    EndIf
    
    If ElapsedMilliseconds() >= lastDelayTime + 10
      Delay(10)
      lastDelayTime = ElapsedMilliseconds()
    EndIf
  Wend

  If p_mode
    ProcedureReturn 1
  Else
    ProcedureReturn maxmaxWrittenSample
  EndIf
EndProcedure

Procedure.l QBPlay_WriteNote(*lpActSample, p_fq.f, noteOnLengthSample.l, *p_WAVfmtEx.WAVEFORMATEX, *p_tone.QBSong_ToneData, p_mix.l)
  Protected actSampleVal.l, maxAmp.l, maxAmpF.f
  Protected fadeInSamples.l, fadeOutSample.l
  Protected writtenSamples.l
  Protected count.l, countStep.l, i.l, i2.l, temp.l, tempF.f, j.l
  Protected resume.l
  
  maxAmpF   = (Pow(2, *p_WAVfmtEx\wBitsPerSample) * *p_tone\volume) / 400.0 ; -6dB Headroom
  maxAmp    = maxAmpF
  countStep = maxAmpF * p_fq / (#QBPlay_SampleRate)
  resume    = 0
  noteOn    = 1
  fadeInSamples  = *p_tone\attack * *p_WAVfmtEx\nSamplesPerSec / 1000
  fadeOutSamples = *p_tone\sustain * *p_WAVfmtEx\nSamplesPerSec / 1000
  
  While i < noteOnLengthSample Or i2 < fadeOutSamples
    ; these are the 'instruments' ;-)
    ; they have to generate a samplevalue in range from -maxAmp[F] to maxAmp[F]
    Select *p_tone\id
      Case #QBPlay_Tone_Sin
        tempF = i / *p_WAVfmtEx\nSamplesPerSec
        actSampleVal = maxAmpF * Sin(2.0 * #PI * tempF * p_fq)
      
      Case #QBPlay_Tone_Tan
        tempF = i / *p_WAVfmtEx\nSamplesPerSec
        actSampleVal = maxAmpF * Tan(2.0 * #PI * tempF * p_fq)
        If actSampleVal > maxAmpF : actSampleVal = maxAmpF
        ElseIf  actSampleVal < -maxAmpF : actSampleVal = -maxAmpF
        EndIf
        actSampleVal / 2.0
         
      Case #QBPlay_Tone_Saw
        actSampleVal = count / 2.0
        count + countStep
        If count > maxAmp
          count = 0 - maxAmp
        EndIf
         
      Case #QBPlay_Tone_Rect
        tempF = i / *p_WAVfmtEx\nSamplesPerSec
        actSampleVal = maxAmpF * Sin(2.0 * #PI * tempF * p_fq)
        If actSampleVal >= 0
          actSampleVal = maxAmp
        Else
          actSampleVal = 0 - maxAmp
        EndIf
        actSampleVal / 4.0
      
      Case #QBPlay_Tone_Dirac
        count = *p_WAVfmtEx\nSamplesPerSec / p_fq
        If i%count = 0
          actSampleVal = maxAmp
        Else
          actSampleVal = 0
        EndIf

      Case #QBPlay_Tone_Organ
        tempF = i / *p_WAVfmtEx\nSamplesPerSec
        actSampleVal = maxAmpF * Sin(2.0 * #PI * tempF * p_fq) / 4
        For j=1 To 3
          tempF = i / *p_WAVfmtEx\nSamplesPerSec * Pow(2, j)
          actSampleVal + maxAmpF * Sin(2.0 * #PI * tempF * p_fq) / 4
        Next

      Case #QBPlay_Tone_ChurchOrgan
        tempF = i / *p_WAVfmtEx\nSamplesPerSec
        actSampleVal = maxAmpF * Sin(2.0 * #PI * tempF * p_fq) / 2
        
        For j=1 To 8
          tempF = i / *p_WAVfmtEx\nSamplesPerSec * Pow(2, j)
          actSampleVal + maxAmpF * Sin(2.0 * #PI * tempF * p_fq) / Pow(2, j + 1)
        Next
         
    EndSelect

    ; correct it by attack/sustain
    If noteOn
      tempF = Sqr(i / fadeInSamples)
      If i < fadeInSamples
        actSampleVal = actSampleVal * tempF
      EndIf
    Else
      tempF = Sqr(i2 / fadeOutSamples)
      actSampleVal = actSampleVal - actSampleVal * tempF
      i2 + 1
    EndIf

    If *p_WAVfmtEx\nChannels <= 2 ; no other than stereo at the moment
      If *p_tone\pan > 0
        actSampleValL = (100 - *p_tone\pan) * actSampleVal / 100
      Else
        actSampleValL = actSampleVal
      EndIf
      If *p_tone\pan < 0
        actSampleValR = (100 + *p_tone\pan) * actSampleVal / 100
      Else
        actSampleValR = actSampleVal
      EndIf

      If p_mix = 0 ; don't keep original
        If *p_WAVfmtEx\wBitsPerSample <= 16
          PokeW(*lpActSample, actSampleValL)
          *lpActSample + 2
          PokeW(*lpActSample, actSampleValR)
          *lpActSample + 2
        EndIf
      Else ; mix with original Data
        If *p_WAVfmtEx\wBitsPerSample <= 16
          ; add left
          actSampleValL = actSampleValL + PeekW(*lpActSample)
          If actSampleValL > 32767
            actSampleValL = 32767
          ElseIf actSampleValL < -32768
            actSampleValL = -32768
          EndIf
          PokeW(*lpActSample, actSampleValL)
          *lpActSample + 2

          ; add right
          actSampleValR = actSampleValR + PeekW(*lpActSample)
          If actSampleValR > 32767
            actSampleValR = 32767
          ElseIf actSampleValR < -32768
            actSampleValR = -32768
          EndIf
          PokeW(*lpActSample, actSampleValR)
          *lpActSample + 2
        EndIf
      EndIf
    EndIf
    
    i + 1

    If i = noteOnLengthSample
      noteOn = 0
    EndIf  
  Wend

  ProcedureReturn i
EndProcedure
 
; IDE Options = PureBasic v4.02 (Windows - x86)
; CursorPosition = 47
; Folding = ---